/*
 * Amiga MUD
 *
 * Copyright (c) 1999 by Chris Gray
 */

/*
 * Common/baseUtil.m - basic utility routines. Otherwise unclassified.
 */

public t_baseUtil CreateTable()$
use t_baseUtil

private tp_baseUtil CreateTable()$
use tp_baseUtil

/*
 * DoChecks - run through an action list, calling each one. Assume they
 *	all return a status value. Stop if one returns other than continue,
 *	and return the overall result. This will be:
 *		fail - a checker returned fail, i.e. do not proceed
 *		succeed - a checker returned succeed, i.e. this check
 *			overrode all others, but proceed with action
 *		continue - nothing special - proceed with action
 *	The distinction between the last two is only needed if the caller
 *	would also do something similar to the checkers (see the use of
 *	'DoRoomLookChecks' in 'EnterRoom' in 'roomUtil.m'.) Otherwise, most
 *	callers will treat the value as yes/no.
 *	Note: we assume that none of the actions modifies the list.
 */

/*
 * Note that none of these routines use 'SetEffectiveToNone'. This is a
 * part of the big security hole involving SysAdmin. SysAdmin should never
 * be used to 'play' on a MUD where untrusted people have been given
 * builder, apprentice or wizard status. Those characters can write code
 * and attach it to rooms, objects, etc. That code, if triggered by
 * SysAdmin, will run as SysAdmin. For this reason, SysAdmin should *never*
 * go into the PlayPen!.
 * Adding SetEffectiveToNone is a problem, because it prevents these
 * routines from doing entirely legitimate activities on behalf of the
 * active character. Character idle code for build description routines
 * is an example. Even SysAdmin can't shut down correctly in that case!
 */

define t_baseUtil proc utility public DoChecks(list action la)status:
    int count, i;
    status st;

    st := continue;
    if la ~= nil then
	count := Count(la);
	i := 0;
	while i < count and st = continue do
	    st := call(la[i], status)();
	    i := i + 1;
	od;
    fi;
    st
corp;

/*
 * DoThingChecks - same, except a thing is passed to each action.
 */

define t_baseUtil proc utility public DoThingChecks(list action la;
	thing th)status:
    int count, i;
    status st;

    st := continue;
    if la ~= nil then
	count := Count(la);
	i := 0;
	while i < count and st = continue do
	    st := call(la[i], status)(th);
	    i := i + 1;
	od;
    fi;
    st
corp;

/*
 * DoStringChecks - same, except each checker is passed a string argument.
 */

define t_baseUtil proc utility public DoStringChecks(list action la;
	string s)status:
    int count, i;
    status st;

    st := continue;
    if la ~= nil then
	count := Count(la);
	i := 0;
	while i < count and st = continue do
	    st := call(la[i], status)(s);
	    i := i + 1;
	od;
    fi;
    st
corp;

/*
 * DoIntChecks - same, except each checker is passed an int argument.
 */

define t_baseUtil proc utility public DoIntChecks(list action la; int n)status:
    int count, i;
    status st;

    st := continue;
    if la ~= nil then
	count := Count(la);
	i := 0;
	while i < count and st = continue do
	    st := call(la[i], status)(n);
	    i := i + 1;
	od;
    fi;
    st
corp;

/*
 * DoList - run through a list of actions, doing all of them.
 *	This routine specifically handles the case of the action
 *	deleting itself from the list.
 */

define t_baseUtil proc utility public DoList(list action la)void:
    int i, count, newCount;

    if la ~= nil then
	count := Count(la);
	i := 0;
	while i < count do
	    call(la[i], void)();
	    newCount := Count(la);
	    if newCount = count then
		i := i + 1;
	    else
		count := newCount;
	    fi;
	od;
    fi;
corp;

/*
 * DoThingActions - these are passed a thing, but return no result.
 */

define t_baseUtil proc utility public DoThingActions(list action la;
	thing th)void:
    int i;

    if la ~= nil then
	for i from 0 upto Count(la) - 1 do
	    call(la[i], void)(th);
	od;
    fi;
corp;

/*
 * DoActionsString - do some actions, yielding a string.
 */

define t_baseUtil proc utility public DoActionsString(list action la)string:
    int i;
    string s;

    s := "";
    if la ~= nil then
	for i from 0 upto Count(la) - 1 do
	    s := s + call(la[i], string)();
	od;
    fi;
    s
corp;

/*
 * addChecker - utility for adding checkers to checker lists.
 */

define tp_common proc utility addChecker(thing th; property list action pla;
	action a; bool front)void:
    list action la;

    la := th@pla;
    if la = nil then
	/* there was no list yet - create one */
	la := CreateActionList();
	th@pla := la;
	AddTail(la, a);
    else
	if front then
	    AddHead(la, a);
	else
	    AddTail(la, a);
	fi;
    fi;
corp;

/*
 * delChecker - utility for deleting checkers and removing the list if
 *	it is now empty.
 */

define tp_common proc utility delChecker(thing th; property list action pla;
	action a)void:
    list action la;

    la := th@pla;
    if la ~= nil then
	DelElement(la, a);
	if Count(la) = 0 then
	    th -- pla;
	fi;
    fi;
corp;


/* routines for inserting/deleting checkers in locations for specific verbs */

define t_baseUtil proc utility AddLocationGetChecker(thing location; action a;
	bool front)void:
    addChecker(location, p_lGetChecks, a, front);
corp;

define t_baseUtil proc utility DelLocationGetChecker(thing location;
	action a)void:
    delChecker(location, p_lGetChecks, a);
corp;

define t_baseUtil proc utility AddLocationDropChecker(thing location; action a;
	bool front)void:
    addChecker(location, p_lDropChecks, a, front);
corp;

define t_baseUtil proc utility DelLocationDropChecker(thing location;
	action a)void:
    delChecker(location, p_lDropChecks, a);
corp;

define t_baseUtil proc utility AddLocationSayChecker(thing location; action a;
	bool front)void:
    addChecker(location, p_lSayChecks, a, front);
corp;

define t_baseUtil proc utility DelLocationSayChecker(thing location;
	action a)void:
    delChecker(location, p_lSayChecks, a);
corp;

define t_baseUtil proc utility AddLocationLookChecker(thing location; action a;
	bool front)void:
    addChecker(location, p_lLookChecks, a, front);
corp;

define t_baseUtil proc utility DelLocationLookChecker(thing location;
	action a)void:
    delChecker(location, p_lLookChecks, a);
corp;

define t_baseUtil proc utility AddLocationLightChecker(thing location;
	action a; bool front)void:
    addChecker(location, p_lLightChecks, a, front);
corp;

define t_baseUtil proc utility DelLocationLightChecker(thing location;
	action a)void:
    delChecker(location, p_lLightChecks, a);
corp;

/* Call up those checker lists. */

define t_baseUtil proc utility DoLocationSayChecks(thing location;
	string said)status:
    DoStringChecks(location@p_lSayChecks, said)
corp;

define t_baseUtil proc utility DoLocationLookChecks(thing location)status:
    DoChecks(location@p_lLookChecks)
corp;

define t_baseUtil proc utility DoLocationLightChecks(thing location)status:
    DoChecks(location@p_lLightChecks)
corp;

define t_baseUtil proc utility DoLocationGetChecks(
	thing location, object)status:
    DoThingChecks(location@p_lGetChecks, object)
corp;

define t_baseUtil proc utility DoLocationDropChecks(
	thing location, object)status:
    DoThingChecks(location@p_lDropChecks, object)
corp;


/*
 * define a unique-id generator for sound-effects id's
 */

define tp_baseUtil EffectIdBase CreateThing(nil)$
define tp_baseUtil eid_nextId CreateIntProp()$
EffectIdBase@eid_nextId := 0$

define t_baseUtil proc NextSoundEffectId()int:

    EffectIdBase@eid_nextId := EffectIdBase@eid_nextId + 1;
    EffectIdBase@eid_nextId
corp;


/* Properties, etc. that control the automated backups and other globally
   settable things. */

define tp_common GlobalThing CreateThing(nil)$
define tp_common p_FlushNeeded CreateBoolProp()$
define tp_common p_BackupInterval CreateIntProp()$

SetThingStatus(GlobalThing, ts_public)$ /* so parseInput can write it */
GlobalThing@p_FlushNeeded := false$
GlobalThing@p_BackupInterval := 0$

define tp_common p_FreePoses CreateBoolProp()$
GlobalThing@p_FreePoses := false$

define tp_common p_FreeBuilding CreateBoolProp()$
GlobalThing@p_FreeBuilding := false$

define tp_common p_AllowShout CreateBoolProp()$
GlobalThing@p_AllowShout := false$


/*
 * CharacterNameS - a standardized form of a specific character name.
 */

define t_baseUtil proc utility public CharacterNameS(thing who)string:
    string name;

    name := FormatName(who@p_pName);
    if who@p_pStandard then
	name
    else
	"the " + name
    fi
corp;

/*
 * CharacterNameG - a standardized form of a generic character name.
 */

define t_baseUtil proc utility public CharacterNameG(thing who)string:
    string name;

    name := FormatName(who@p_pName);
    if who@p_pStandard then
	name
    else
	AAn("", name)
    fi
corp;


/*
 * InsertCommand - do a command, but display it first.
 */

define t_baseUtil proc utility public InsertCommand(string command)void:

    Print(">>> " + command + "\n");
    GlobalThing@p_FlushNeeded := true;
    ignore Parse(G, command);
corp;


/* This is used to allow creation of a carrying list on machines - it does
   it carefully. */

define t_baseUtil proc utility public SetupMachine(thing machine)void:

    if machine@p_pCarrying = nil then
	machine@p_pCarrying := CreateThingList();
    fi;
    if machine@p_pHiddenList = nil then
	machine@p_pHiddenList := CreateThingList();
    fi;
    if machine@p_pEnterActions = nil then
	machine@p_pEnterActions := CreateActionList();
    fi;
    if machine@p_pExitActions = nil then
	machine@p_pExitActions := CreateActionList();
    fi;
corp;


/*
 * FakeModel - setup a simple object.
 */

define t_baseUtil proc utility public FakeModel(thing object;
	string name, desc)void:

    object@p_oName := name;
    if desc ~= "" then
	object@p_oDesc := desc;
    fi;
    object@p_oNotGettable := true;
    object@p_oInvisible := true;
    SetThingStatus(object, ts_wizard);
corp;


/* a couple of handy utilities */

define t_baseUtil proc IsYes(string s)bool:

    s == "y" or s == "yes" or s == "t" or s == "true"
corp;

define t_baseUtil proc IsNo(string s)bool:

    s == "n" or s == "no" or s == "n" or s == "false"
corp;

/*
 * ShowList - print out a contents/carrying list. Return 'true' if nothing
 *	was printed.
 */

define t_baseUtil proc utility public ShowList(list thing lt;
	string starter)bool:
    int i;
    thing object;
    string s;
    bool first;

    first := true;
    if lt ~= nil then
	for i from 0 upto Count(lt) - 1 do
	    object := lt[i];
	    if not object@p_oInvisible then
		if first then
		    first := false;
		    Print(starter);
		fi;
		Print("  " + FormatName(object@p_oName) + "\n");
	    fi;
	od;
    fi;
    first
corp;

/*
 * DoAll - do the given proc for each visible thing in the given list. Return
 *	'continue' if at least one is done and all done yield 'true';
 *	return 'succeed' if one does not return 'true', and return 'fail'
 *	if there are none to do.
 */

define t_baseUtil proc public DoAll(list thing lt; action a)status:
    int count, i, oldCount;
    thing th;
    bool ok, doneOne;

    doneOne := false;
    if lt ~= nil then
	count := Count(lt);
	i := 0;
	ok := true;
	while ok and i ~= count do
	    th := lt[i];
	    if not th@p_oInvisible then
		doneOne := true;
		if call(a, bool)(th) then
		    oldCount := count;
		    count := Count(lt);
		    i := i - (oldCount - count) + 1;
		else
		    ok := false;
		fi;
	    else
		i := i + 1;
	    fi;
	od;
    fi;
    if doneOne then
	if ok then continue else succeed fi
    else
	fail
    fi
corp;


/*
 * HasLight - return 'true' if the given thing (character) supplies light.
 */

define t_baseUtil proc public HasLight(thing who)bool:

    who@p_oLight or FindFlagOnList(who@p_pCarrying, p_oLight)
corp;

/* NOTE: When a monster is killed, DoDrop is called to drop whatever it is
   carrying. That in turn calls CanSee, with 'thePlayer' set to the monster. */

define t_baseUtil proc public CanSee(thing location, thePlayer)bool:

    if location = nil then
	/* Saw this once in log file - something to do with an absent agent. */
	false
    elif not location@p_lDark then
	true
    else
	if thePlayer ~= nil and not thePlayer@p_pStandard then
	    thePlayer := Me();
	fi;
	if thePlayer = nil then
	    location@p_lContents ~= nil and
		FindFlagOnList(location@p_lContents, p_oLight) or
	    FindAgentWithFlag(location, p_oLight) ~= nil or
	    FindAgentWithFlagOnList(location, p_pCarrying, p_oLight) ~= nil
	elif thePlayer@p_oLight then
	    true
	else
	    FindFlagOnList(thePlayer@p_pCarrying, p_oLight) or
	    location@p_lContents ~= nil and
		FindFlagOnList(location@p_lContents, p_oLight) or
	    FindAgentWithFlag(location, p_oLight) ~= nil or
	    FindAgentWithFlagOnList(location, p_pCarrying, p_oLight) ~= nil
	fi
    fi
corp;

/*
 * LightAt - return 'true' if there is light in the given location, without
 *	the current player being considered to be there.
 */

define t_baseUtil proc public LightAt(thing location)bool:

    not location@p_lDark or
    location@p_lContents ~= nil and
	FindFlagOnList(location@p_lContents, p_oLight) or
    FindAgentWithFlag(location, p_oLight) ~= nil or
    FindAgentWithFlagOnList(location, p_pCarrying, p_oLight) ~= nil
corp;

/*
 * ShowHere - generic utility, which works in all modes, to show
 *	an activity to characters nearby.
 *
 *  This routine is in the public table since it is used by the build code.
 */

public proc public ShowHere(string what)void:

    call(Me()@p_pCurrentMode@p_lmShowOthersHere, void)(what);
corp;

/*
 * MaybeShowHere - generic utility, which works in all modes, to show
 *	an activity to characters nearby, depending on visibility.
 */

define t_baseUtil proc public MaybeShowHere(string what)void:
    thing me, lm;

    me := Me();
    lm := me@p_pCurrentMode;
    if not me@p_pHidden and call(lm@p_lmCanSeeHere, bool)(me) then
	call(lm@p_lmShowOthersHere, void)(what);
    fi;
corp;

/*
 * MaybeShowAlt - variant that accepts a "do-er", and an alternate
 *	string if can't see.
 */

define t_baseUtil proc MaybeShowAlt(thing who; string what, alt)void:
    thing lm;

    lm := who@p_pCurrentMode;
    if not who@p_pHidden and call(lm@p_lmCanSeeHere, bool)(who) then
	call(lm@p_lmShowOthersHere, void)(what);
    else
	call(lm@p_lmShowOthersHere, void)(alt);
    fi;
corp;

/*
 * MaybeShowMeDo - variant that prefixes with character name.
 */

define t_baseUtil proc public MaybeShowMeDo(string what)void:
    thing me, lm;

    me := Me();
    lm := me@p_pCurrentMode;
    if not me@p_pHidden and call(lm@p_lmCanSeeHere, bool)(me) then
	call(lm@p_lmShowOthersHere, void)(
	    Capitalize(CharacterNameS(me)) + " " + what + ".\n");
    fi;
corp;


/*
 * Scenery - add some names that will be found by 'look', etc., but that
 *	have nothing special about them. Doing it this way saves creating
 *	a lot of extra things.
 */

define t_baseUtil proc utility public Scenery(thing location;
	string newScenery)void:
    string oldScenery;
    int len;

    len := Length(newScenery);
    if len ~= 0 then
	len := len - 1;
	/* Trim any extra final '.' given. */
	if SubString(newScenery, len, 1) = "." then
	    newScenery := SubString(newScenery, 0, len);
	fi;
	oldScenery := location@p_lScenery;
	if oldScenery ~= "" then
	    newScenery := oldScenery + "." + newScenery;
	fi;
	location@p_lScenery := newScenery;
    fi;
corp;

/*
 * LocationName - set the graphics name of the location.
 */

define t_baseUtil proc public utility LocationName(thing location;
    string name1, name2)void:

    location@p_lName1 := name1;
    if name2 ~= "" then
	location@p_lName2 := name2;
    else
	location -- p_lName2;
    fi;
corp;


/*
 * CarryingChild - return any child of the given thing that the given
 *	player is carrying, either directly or indirectly.
 */

define tp_baseUtil proc containsChild(list thing lt; thing what)thing:
    int count, i;
    thing th;

    if lt = nil then
	nil
    else
	count := Count(lt);
	i := 0;
	while i < count do
	    th := lt[i];
	    if Parent(th) = what then
		i := count + 1;
	    elif th@p_oContents ~= nil then
		th := containsChild(th@p_oContents, what);
		if th ~= nil then
		    i := count + 1;
		else
		    i := i + 1;
		fi;
	    else
		i := i + 1;
	    fi;
	od;
	if i > count then th else nil fi
    fi
corp;

define t_baseUtil proc CarryingChild(thing who, what)thing:
    containsChild(who@p_pCarrying, what)
corp;

/*
 * ChildLocation - return any child of the given thing that the given location
 *	contains, either directly or indirectly.
 */

define t_baseUtil proc ChildLocation(thing location, what)thing:
    containsChild(location@p_lContents, what)
corp;

/*
 * SayToList - output a given message to each location on a list.
 */

define t_baseUtil proc SayToList(list thing lt; string message)void:
    int i;

    if lt ~= nil then
	for i from 0 upto Count(lt) - 1 do
	    ABPrint(lt[i], nil, nil, message);
	od;
    fi;
corp;



/***************************************************************************/

/*
 * GetDocument - get a long document - e.g. description, letter, etc.
 *	Callable by player or machine - will do nothing for a machine.
 *	There are some things to watch out for here. We want these routines
 *	to be 'utility' so that they can be used properly by the build code.
 */

define tp_baseUtil p_pOldDoc CreateStringProp()$
define tp_baseUtil p_pEndAction CreateActionProp()$
define tp_baseUtil p_pRawDocument CreateBoolProp()$
define tp_baseUtil p_pTempString CreateStringProp()$
define tp_baseUtil p_pSavePrompt CreateStringProp()$
define tp_baseUtil p_pSaveAction CreateActionProp()$

/* 'docReset' is replaced later, when its references can be satisfied */
define tp_baseUtil proc docReset()void: corp;

/* Append the given string to the document being built up. This was
   originally set up for the line-by-line mode, but is also used for
   its handling of the end-of-input condition (s = ".") */
define tp_baseUtil proc utility appendToDocument(string line)void:
    action endAction;
    thing me, letter;
    string s;
    int len;

    me := Me();
    s := me@p_pTempString;
    len := Length(s);
    if line = "." or line = "$" then
	if len >= 4000 then
	    Print("*** Warning - input may have been truncated. ***\n");
	fi;
	endAction := me@p_pEndAction;
	docReset();
	/* call this so that, e.g. a normal person using the build code can
	   modify his own objects that are ts_readonly */
	call(endAction, void)(s);
    else
	if len >= 4000 then
	    Print("*** Warning - input has been truncated. ***\n");
	else
	    if me@p_pRawDocument then
		me@p_pTempString := s + line + "\n";
	    else
		if line ~= "" then
		    if s ~= "" then
			s := s + " ";
		    fi;
		    me@p_pTempString := s + line;
		fi;
	    fi;
	fi;
    fi;
corp;

/* Called when the player exits MUD while still entering a document. */
define tp_baseUtil proc utility docIdleAction()void:

    Me()@p_pTempString := Me()@p_pOldDoc;
    appendToDocument(".");
corp;

/* Called by 'activeAction' if the database was saved while the player
   was entering a document, and the server was later restarted from that
   saved database. */
define tp_baseUtil proc docActiveAction()void:

    Print("\n* You were editing/entering something when the last backup "
	  "was made. That text change has been lost. *\n\n");
    docIdleAction();
corp;

/* We can now define the full body of this. Basically, just clean up after
   doing a GetDocument. */
replace docReset()void:
    thing me;

    me := Me();
    DelElement(me@p_pEnterActions, docActiveAction);
    DelElement(me@p_pExitActions, docIdleAction);
    if me@p_pSaveAction ~= nil then
	/* not setup if using 'EditString' */
	ignore SetCharacterInputAction(me@p_pSaveAction);
	me -- p_pSaveAction;
	ignore SetPrompt(me@p_pSavePrompt);
	me -- p_pSavePrompt;
    fi;
    me -- p_pTempString;
    me -- p_pOldDoc;
    me -- p_pEndAction;
    me -- p_pRawDocument;
corp;

/* Called when the user finished editting the document, as part of
   'EditString'. */
define tp_baseUtil proc utility docEndAction(string s; bool ok)void:
    thing me;

    me := Me();
    if ok then
	me@p_pTempString := s;
    else
	me@p_pTempString := me@p_pOldDoc;
    fi;
    appendToDocument(".");
corp;

/* Set things up so that user input will be accumulated into a document
   of some kind, which will then be passed to 'endAction'. Also handle
   cases of input not completed when player exits, or when a backup is
   made and later used. */
define t_baseUtil proc utility GetDocument(string prompt, intro, oldDoc;
	action endAction; bool isRaw)bool:
    thing me;
    action oldAction;

    me := Me();
    if CanEdit() then
	if Editing() then
	    Print("You are alreadying editing something!\n");
	    false
	else
	    me@p_pTempString := "";
	    me@p_pEndAction := endAction;
	    me@p_pRawDocument := isRaw;
	    me@p_pOldDoc := oldDoc;
	    AddHead(me@p_pEnterActions, docActiveAction);
	    AddHead(me@p_pExitActions, docIdleAction);
	    EditString(oldDoc, docEndAction,
		       if isRaw then EDIT_CODE else EDIT_COOKED fi,
		       intro);
	    true
	fi
    else
	oldAction := SetCharacterInputAction(appendToDocument);
	if oldAction = nil then
	    /* must have been a machine! */
	    OPrint(Capitalize(CharacterNameS(me)) + " is confused.\n");
	    false
	else
	    me@p_pTempString := "";
	    me@p_pEndAction := endAction;
	    me@p_pRawDocument := isRaw;
	    me@p_pOldDoc := oldDoc;
	    AddHead(me@p_pEnterActions, docActiveAction);
	    AddHead(me@p_pExitActions, docIdleAction);
	    me@p_pSavePrompt := SetPrompt("* " + prompt);
	    me@p_pSaveAction := oldAction;
	    Print(intro +
		". End with a line containing only a single period.\n");
	    true
	fi
    fi
corp;

/*
 * GetCheckedDescription - variant which lets a routine of the callers handle
 *	each of the input lines. Note that since we need line-by-line
 *	processing, there is no 'EditString' variation here.
 */

define t_baseUtil proc utility GetCheckedEnd()void: corp;  /* replaced later */

define tp_baseUtil proc checkedActiveAction()void:

    Print("\n* You were entering something when the last backup "
	  "was made. Those changes have been lost. *\n\n");
    GetCheckedEnd();
corp;

/*
 * GetCheckedEnd - this can be called by other code (e.g. the handler for
 *	this checked description) to terminate this input mode, e.g.
 *	because of erroneous user input.
 */

replace GetCheckedEnd()void:
    thing me;

    me := Me();
    ignore SetCharacterInputAction(me@p_pSaveAction);
    me -- p_pSaveAction;
    ignore SetPrompt(me@p_pSavePrompt);
    me -- p_pSavePrompt;
    DelElement(me@p_pEnterActions, checkedActiveAction);
    DelElement(me@p_pExitActions, GetCheckedEnd);
corp;

define t_baseUtil proc utility GetCheckedDescription(string prompt;
	action lineHandler)bool:
    thing me;
    action oldAction;

    me := Me();
    oldAction := SetCharacterInputAction(lineHandler);
    if oldAction = nil then
	/* must have been a machine! */
	OPrint(Capitalize(CharacterNameS(me)) + " is confused.\n");
	false
    else
	me@p_pSaveAction := oldAction;
	me@p_pSavePrompt := SetPrompt("* " + prompt);
	AddHead(me@p_pEnterActions, checkedActiveAction);
	AddHead(me@p_pExitActions, GetCheckedEnd);
	true
    fi
corp;


/***************************************************************************/

/*
 * Paginate - paginate a string within the output screen size.
 */

define tp_baseUtil p_pPaginateSetup CreateBoolProp()$
define tp_baseUtil p_pPaginateString CreateStringProp()$
define tp_baseUtil p_pPaginateLen CreateIntProp()$
define tp_baseUtil p_pPaginatePrompt CreateStringProp()$
define tp_baseUtil p_pPaginateHandler CreateActionProp()$

define tp_baseUtil proc paginateReset()void:
    thing me;

    me := Me();
    me -- p_pPaginateString;
    me -- p_pPaginateLen;
    if me@p_pPaginateSetup then
	me -- p_pPaginateSetup;
	DelElement(me@p_pEnterActions, paginateReset);
	DelElement(me@p_pExitActions, paginateReset);
	ignore SetPrompt(me@p_pPaginatePrompt);
	me -- p_pPaginatePrompt;
	ignore SetCharacterInputAction(me@p_pPaginateHandler);
	me -- p_pPaginateHandler;
    fi;
corp;

define tp_baseUtil proc paginateParse(string line)void: corp;

define tp_baseUtil proc paginateShowPage()void:
    thing me;
    string s;
    int len, i, line, height, width;

    me := Me();
    s := me@p_pPaginateString;
    len := me@p_pPaginateLen;
    height := TextHeight(0) - 1;
    width := TextWidth(0);
    line := 1;
    while len > 0 and line <= height do
	i := Index(s, "\n");
	if i = -1 then
	    Print(s);
	    Print("\n");
	    len := 0;
	else
	    i := i + 1;
	    Print(SubString(s, 0, i));
	    s := SubString(s, i, len - i);
	    len := len - i;
	    if i <= width then
		line := line + 1;
	    else
		line := line + (i + width - 4) / (width - 9);
	    fi;
	fi;
    od;
    if len = 0 then
	paginateReset();
    else
	me@p_pPaginateString := s;
	me@p_pPaginateLen := len;
	if not me@p_pPaginateSetup then
	    me@p_pPaginateSetup := true;
	    me@p_pPaginatePrompt := SetPrompt("[M O R E] ");
	    me@p_pPaginateHandler := SetCharacterInputAction(paginateParse);
	    AddHead(me@p_pEnterActions, paginateReset);
	    AddHead(me@p_pExitActions, paginateReset);
	fi;
    fi;
corp;

replace paginateParse(string line)void:

    if line = "" then
	paginateShowPage();
    elif line == "q" then
	paginateReset();
    else
	Print("Options are:\n  q - quit\n  empty line - next page\n");
    fi;
corp;

define t_baseUtil proc Paginate(string s)void:
    thing me;
    int len;

    len := Length(s);
    if SubString(s, len - 1, 1) ~= "\n" then
	s := s + "\n";
	len := len + 1;
    fi;
    me := Me();
    if GType(nil) == "amiga" or GType(nil) == "java" then
	/* using full MUD client - no reason to paginate */
	Print(s);
    elif Character(me@p_pName) = nil then
	OPrint(Capitalize(CharacterNameS(me)) + " is confused.\n");
    else
	me@p_pPaginateString := s;
	me@p_pPaginateLen := len;
	paginateShowPage();
    fi;
corp;


/*
 * ShowPosition - return the position of the given player/character.
 */

define t_baseUtil proc public ShowPosition(thing th)string:
    int pos;

    pos := th@p_pPosition;
    if pos ~= POS_NONE then
	case pos
	incase POS_SIT_IN:
	    "sitting in"
	incase POS_SIT_ON:
	    "sitting on"
	incase POS_LIE_IN:
	    "lying in"
	incase POS_LIE_ON:
	    "lying on"
	incase POS_STAND_IN:
	    "standing in"
	incase POS_STAND_ON:
	    "standing on"
	default:
	    "relative to"
	esac +
	" the " + FormatName(th@p_pWhere@p_oName) + "."
    else
	"here."
    fi
corp;

/*
 * ZapObject - destroy an object, and anything it contains.
 */

define t_baseUtil proc public ZapObject(thing object)void:
    list thing lt;
    int count;
    thing th;

    CancelAllDoAfters(object);
    lt := object@p_oContents;
    if lt ~= nil then
	count := Count(lt);
	while count ~= 0 do
	    count := count - 1;
	    th := lt[count];
	    ZapObject(th);
	    DelElement(lt, th);
	od;
    fi;
    ClearThing(object);
corp;

/*
 * SaySubject - little utility for machines - check if the given word is
 *	any of the common forms of address for the given name.
 */

define t_baseUtil proc SaySubject(string word, name)bool:

    word == name or word == name + "," or word == name + ":"
corp;

/*
 * CarryItem - try to add an item to the carry list. If too many complain
 *	and return false, else add the item and return true.
 */

define t_baseUtil proc CarryItem(thing object)bool:
    thing me;
    list thing carrying;

    me := Me();
    carrying := me@p_pCarrying;
    if Count(carrying) >= MAX_CARRY then
	Print("You can't carry anything else.\n");
	false
    else
	AddTail(carrying, object);
	object@p_oCarryer := me;
	true
    fi
corp;

/*
 * LookAtObject - common code to do the looking at a normal object.
 */

define t_baseUtil proc LookAtObject(thing object)bool:
    status st;
    action a;
    string s;
    int saveComponent;

    st := continue;
    a := object@p_oDescCheck;
    if a ~= nil then
	st := call(a, status)();
    fi;
    if st ~= fail then
	a := object@p_oDescAction;
	if a = nil then
	    s := object@p_oDesc;
	    if s = "" then
		Print("You see nothing special about the " +
		    FormatName(object@p_oName) + ".\n");
	    else
		NPrint(s);
	    fi;
	else
	    SetIt(object);
	    Print(call(a, string)() + "\n");
	fi;
	if object@p_Image ~= "" then
	    ShowImage(object@p_Image);
	    a := Me()@p_pCurrentMode@p_lmClearMap;
	    if a ~= nil then
		call(a, void)();
	    fi;
	elif object@p_oDrawAction ~= nil then
	    ShowDrawing(object@p_oDrawAction, object);
	    a := Me()@p_pCurrentMode@p_lmClearMap;
	    if a ~= nil then
		call(a, void)();
	    fi;
	fi;
	true
    else
	false
    fi
corp;


/* Also used in the bulletin board code */
define t_baseUtil proc ReadObject(thing object)bool:
    action a;

    SetIt(object);
    a := object@p_oReadAction;
    if a ~= nil then
	Paginate(call(a, string)());
    elif object@p_oReadString ~= "" then
	Paginate(object@p_oReadString);
    else
	Print("There is nothing to read on the " +
	    FormatName(object@p_oName) + ".\n");
    fi;
    true
corp;

define tp_baseUtil proc doReadObject(thing object)bool:

    ignore ReadObject(object);
    true
corp;


/*
 * DoSay - bottom level of saying - here since 'parseInput' uses it.
 *	Needs to be not 'utility' in order to use 'Say'.
 */

define t_baseUtil proc public DoSay(string what)void:
    thing me, lm, location;

    while SubString(what, 0, 1) = " " do
	what := SubString(what, 1, Length(what) - 1);
    od;
    me := Me();
    lm := me@p_pCurrentMode;
    location := call(lm@p_lmGetLocation, thing)();
    if location = nil or DoLocationSayChecks(location, what) = continue then
	if me@p_pEchoPose then
	    Print("You say: " + what + "\n");
	fi;
	call(lm@p_lmSayOthersHere, void)(
	    if me@p_pHidden or not call(lm@p_lmCanSeeHere, bool)(me) then
	    "Someone" else "" fi, what);
    fi;
corp;

/*
 * checkAlias - check for and handle an alias on the character. This is
 *	a separate routine so that it is NOT 'utility', but is owned by
 *	SysAdmin. This lets folks other than SysAdmin read the alias
 *	things!
 */

define tp_baseUtil proc checkAlias(string s)string:
    list thing aliases;
    thing alias;
    string word;
    int count;
    bool doneIt;

    SetTail(s);
    word := GetWord();
    aliases := Me()@p_pAliases;
    if aliases ~= nil then
	count := Count(aliases);
	doneIt := false;
	while count ~= 0 and not doneIt do
	    count := count - 1;
	    alias := aliases[count];
	    if alias@p_sAliasKey == word then
		s := alias@p_sAliasValue + " " + GetTail();
		doneIt := true;
	    fi;
	od;
    fi;
    s
corp;

/*
 * AddSpecialCommand - set things up to add a special this-location-only
 *	command. Note: allowing the use of this procedure is a fairly large
 *	security hole, since the commands thus added take precedence OVER
 *	the normal ones, thus this routine can be used to invoke arbitrary
 *	action when a player uses a presumed safe command. See 'parseInput'
 *	for the actual use of these values.
 */

define tp_baseUtil p_lSpecialWords CreateStringProp()$
define tp_baseUtil p_lSpecialActions CreateActionListProp()$

define t_baseUtil proc utility public AddSpecialCommand(thing location;
	string command; action a)void:
    list action la;

    la := location@p_lSpecialActions;
    if la = nil then
	la := CreateActionList();
	location@p_lSpecialActions := la;
    fi;
    location@p_lSpecialWords := location@p_lSpecialWords + command + ".";
    AddTail(la, a);
corp;

define t_baseUtil proc utility public RemoveSpecialCommand(thing location;
	string command; action a)bool:
    list action la;
    string s;
    int pos, len;

    la := location@p_lSpecialActions;
    if la ~= nil then
	s := location@p_lSpecialWords;
	command := command + ".";
	pos := Index(s, command);
	if pos ~= -1 then
	    len := Length(command);
	    location@p_lSpecialWords := SubString(s, 0, pos) +
		SubString(s, pos + len, Length(s) - pos - len);
	    DelElement(la, a);
	    true
	else
	    false
	fi
    else
	false
    fi
corp;

/*
 * parseInput - the normal input command handler.
 *	NOTE: we want this to be 'utility', so that the build commands can
 *	be run by the real player. The problem with this is that any message
 *	from 'Parse' will have an '@' in front if it is run by a non-wizard.
 *	I'll kluge that by having 'Parse' force wizard-mode if it needs to
 *	print an error message. Sigh.
 *	Another '@' problem surfaces: if the whole result of the user command
 *	is to print a 'p_oActString', then that output will be prefixed with
 *	'@' when a non-wizard is running. I have implemented 'NPrint' to
 *	hopefully get around this.
 */

define tp_common p_pWizardAction CreateActionProp()$

define tp_common proc utility parseInput(string s)void:
    action a;
    list thing lt;
    thing lm, location, it, me;
    string word, specials;
    int which, count;
    bool doneIt;

    while SubString(s, 0, 1) = ";" do
	s := SubString(s, 1, Length(s) - 1);
    od;
    if s ~= "" then
	me := Me();
	GlobalThing@p_FlushNeeded := true;
	word := SubString(s, 0, 1);
	if word = "." and IsWizard() then
	    me@p_pWizardAction :=
		StringToAction(SubString(s, 1, Length(s) - 1));
	    call(me@p_pWizardAction, void)();
	    me -- p_pWizardAction;
	elif word = "\"" or word = "'" then
	    DoSay(SubString(s, 1, Length(s) - 1));
	else
	    me := Me();
	    lm := me@p_pCurrentMode;
	    lt := call(lm@p_lmGetObjectsHere, list thing)();
	    s := checkAlias(s);
	    SetTail(s);
	    word := GetWord();
	    if FindName(me@p_pCarrying, p_oActWord, word) = fail and
		(lt = nil or FindName(lt, p_oActWord, word) = fail)
	    then
		location := call(lm@p_lmGetLocation, thing)();
		doneIt := false;
		if location ~= nil then
		    specials := location@p_lSpecialWords;
		    if specials ~= "" then
			which := MatchName(specials, word);
			if which >= 0 then
			    call(location@p_lSpecialActions[which], void)();
			    doneIt := true;
			fi;
		    fi;
		fi;
		if not doneIt then
		    /* most commands are done right here */
		    ignore Parse(G, s);
		fi;
	    else
		it := FindResult();
		a := it@p_oActAction;
		if a ~= nil then
		    SetIt(it);
		    call(a, void)();
		else
		    NPrint(it@p_oActString);
		fi;
	    fi;
	fi;
    fi;
corp;

/*
 * handleRawKey - handle a raw special key-hit.
 *	We want this routine NOT utility, so that there is not an '@' in
 *	front of the command when 'InsertCommand' prints it.
 */

define tp_common proc handleRawKey(int n)void:

    case n
    incase KEY_HELP:
	InsertCommand("help");
    incase KEY_KP_UL:
	InsertCommand("northwest");
    incase KEY_KP_U:
	InsertCommand("north");
    incase KEY_KP_UR:
	InsertCommand("northeast");
    incase KEY_KP_L:
	InsertCommand("west");
    incase KEY_KP_C:
	InsertCommand("look");
    incase KEY_KP_R:
	InsertCommand("east");
    incase KEY_KP_DL:
	InsertCommand("southwest");
    incase KEY_KP_D:
	InsertCommand("south");
    incase KEY_KP_DR:
	InsertCommand("southeast");
    incase KEY_KP_PLUS:
	InsertCommand("up");
    incase KEY_KP_MINUS:
	InsertCommand("down");
    esac;
corp;

/*
 * idleAction - the action that is executed when the player leaves.
 *	These are NOT utility - do not want '@' in front of any output.
 */

define tp_common proc idleAction()void:

    DoList(Me()@p_pExitActions);
    call(Me()@p_pCurrentMode@p_lmIdleAction, void)();
    GlobalThing@p_FlushNeeded := true;
corp;

/*
 * activeAction - the action that is executed when the player re-enters.
 */

define tp_common proc activeAction()void:
    thing me;

    me := Me();
    call(me@p_pCurrentMode@p_lmActiveAction1, void)();
    DoList(me@p_pEnterActions);
    call(me@p_pCurrentMode@p_lmActiveAction2, void)();
corp;


/*
 * RemoveAllFromInventory - remove all occurrences of the given object
 *	from the given characters inventory, including inside containers.
 */

define tp_baseUtil proc scanList(thing who; list thing lt;
	thing what; bool top)void:
    int count, i;
    thing th;

    count := Count(lt);
    i := 0;
    while i ~= count do
	th := lt[i];
	if Parent(th) = what then
	    if top then
		SPrint(who, FormatName(what@p_oName) + " vanishes.\n");
	    fi;
	    ClearThing(th);
	    DelElement(lt, th);
	    count := count - 1;
	else
	    if th@p_oContents ~= nil then
		scanList(who, th@p_oContents, what, false);
	    fi;
	    i := i + 1;
	fi;
    od;
corp;
	
define t_baseUtil proc RemoveAllFromInventory(thing who, what)void:

    scanList(who, who@p_pCarrying, what, true);
corp;


/***************************************************************************/

/* Some general routines for setting up verbs that do things to things.
   'VerbCarry' requires that the player be carrying the object in order to
   do whatever to it. 'VerbHere' allows it to be either carried or in the
   room the player is in. It would also be possible to look at things that
   are carried by other players/machines, but I chose not to. Note that
   I check the player, then the room, then the specific object.
   Return 'false' if we were not able to do the action on the requested
   thing, because the thing is not available, or the action fails.
*/

/* A problem has cropped up with the drinking monsters. The code below
   looks for the same properties on the player, the room and the object.
   So, if you try to drink the drinking troll, it will execute the drink
   action on the troll and say that you shouldn't do that. Unfortunately,
   that action is also done when the troll's special action gets the troll
   to 'drink water'. Proper solution is two more properties, sigh. This
   has been done for the indirect case with 'actorCheck'. Thus we do not
   allow the case of attaching a string to the player which is the entire
   result of trying to do that action. */

define tp_baseUtil proc public commonVerbTail(property string direct;
	property action indirect, actorCheck; thing object;
	string failHeader, verbName, name)bool:
    thing me, location;
    action a;
    string directString;
    status st;
    bool doneOne;

    /* Note: the status values returned by the handler routines are
       interpreted as follows:
	  continue - nothing special - keep looking for something special
	  succeed - successfully handled this case
	  fail - this case is handled, but cease cases and parsing
       The presence of a 'direct' string property is taken to be the same
       as a routine which prints that string and returns 'succeed', with
       the exception that something on a given object will override a
       direct string on a location.
    */

    me := Me();
    location := call(me@p_pCurrentMode@p_lmGetLocation, thing)();
    doneOne := false;
    if actorCheck ~= nil then
	a := me@actorCheck;
	if a ~= nil then
	    SetIt(object);
	    st := call(a, status)();
	    if st ~= continue then
		doneOne := true;
	    fi;
	fi;
    fi;
    if not doneOne and indirect ~= nil and location ~= nil then
	a := location@indirect;
	if a ~= nil then
	    SetIt(object);
	    st := call(a, status)();
	    if st ~= continue then
		doneOne := true;
	    fi;
	fi;
    fi;
    if not doneOne and direct ~= nil and location ~= nil then
	directString := location@direct;
	if directString ~= "" and
	    (object = nil or object@direct = "" and object@indirect = nil)
	then
	    doneOne := true;
	    Print(directString + "\n");
	    st := continue;
	fi;
    fi;
    if not doneOne and object ~= nil and indirect ~= nil then
	a := object@indirect;
	if a ~= nil then
	    SetIt(object);
	    st := call(a, status)();
	    if st ~= continue then
		doneOne := true;
	    fi;
	fi;
    fi;
    if not doneOne and object ~= nil and direct ~= nil then
	directString := object@direct;
	if directString ~= "" then
	    doneOne := true;
	    Print(directString + "\n");
	    st := continue;
	fi;
    fi;
    if doneOne then
	st ~= fail
    elif object = nil then
	Print("You must specify what you want to " + verbName + ".\n");
	false
    else
	Print(failHeader + " the " + name + ".\n");
	true
    fi
corp;

define t_baseUtil proc public VerbCarry(string verbName;
	property string direct; property action indirect, actorCheck;
	string failHeader, what)bool:
    thing object;
    string name;
    status st;
    bool done, ok;
    list thing lt;
    int i, count, oldCount;

    done := false;
    if what = "" then
	object := nil;
    elif what == "all" then
	lt := Me()@p_pCarrying;
	count := Count(lt);
	i := 0;
	ok := true;
	while ok and i ~= count do
	    object := lt[i];
	    if not object@p_oInvisible then
		done := true;
		if commonVerbTail(direct, indirect, actorCheck,
		    object, failHeader, verbName, FormatName(object@p_oName))
		then
		    oldCount := count;
		    count := Count(lt);
		    i := i - (oldCount - count) + 1;
		else
		    ok := false;
		fi;
	    else
		i := i + 1;
	    fi;
	od;
	if not done then
	    done := true;
	    Print("You are not carrying anything obvious to " + verbName +
		  ".\n");
	    ok := false;
	fi;
    else
	name := FormatName(what);
	st := FindName(Me()@p_pCarrying, p_oName, what);
	if st = fail then
	    Print(AAn("You are not carrying", name) + ".\n");
	    ok := false;
	    done := true;
	elif st = continue then
	    Print(Capitalize(name) + " is ambiguous here.\n");
	    ok := false;
	    done := true;
	else
	    object := FindResult();
	    name := FormatName(object@p_oName);
	fi;
    fi;
    if done then
	ok
    else
	commonVerbTail(direct, indirect, actorCheck,
		       object, failHeader, verbName, name)
    fi
corp;

define t_baseUtil proc public VerbHere(string verbName; property string direct;
	property action indirect, actorCheck; string failHeader, what)bool:
    thing me, lm, object;
    list thing lt;
    int count, i, oldCount;
    string ambig, name;
    status st;
    bool done, ok;

    me := Me();
    lm := me@p_pCurrentMode;
    done := false;
    object := nil;
    if what == "all" then
	lt := me@p_pCarrying;
	count := Count(lt);
	i := 0;
	ok := true;
	while ok and i ~= count do
	    object := lt[i];
	    if not object@p_oInvisible then
		done := true;
		if commonVerbTail(direct, indirect, actorCheck, object,
		    failHeader, verbName, FormatName(object@p_oName))
		then
		    oldCount := count;
		    count := Count(lt);
		    i := i - (oldCount - count) + 1;
		else
		    ok := false;
		fi;
	    else
		i := i + 1;
	    fi;
	od;
	lt := call(lm@p_lmGetObjectsHere, list thing)();
	if lt = nil then
	    count := 0;
	else
	    count := Count(lt);
	fi;
	i := 0;
	while ok and i ~= count do
	    object := lt[i];
	    if not object@p_oInvisible then
		done := true;
		if commonVerbTail(direct, indirect, actorCheck,
		    object, failHeader, verbName, FormatName(object@p_oName))
		then
		    oldCount := count;
		    count := Count(lt);
		    i := i - (oldCount - count) + 1;
		else
		    ok := false;
		fi;
	    else
		i := i + 1;
	    fi;
	od;
	if not done then
	    done := true;
	    Print("There is nothing obvious here to " + verbName + ".\n");
	    ok := false;
	fi;
    elif what ~= "" then
	name := FormatName(what);
	ambig := " is ambiguous here.\n";
	st := FindName(me@p_pCarrying, p_oName, what);
	if st = fail then
	    lt := call(lm@p_lmGetObjectsHere, list thing)();
	    if lt = nil then
		st := fail;
	    else
		st := FindName(lt, p_oName, what);
	    fi;
	    if st = fail then
		object := FindAgent(what);
		if object = nil then
		    lt := call(lm@p_lmGetForSaleHere, list thing)();
		    if lt ~= nil and FindName(lt, p_oName, what) ~= fail then
			done := true;
			ok := false;
			Print("You should buy the " + name +
			    " before you try to " + verbName + " it.\n");
		    elif call(lm@p_lmIsSceneryHere, bool)(what) then
			done := true;
			ok := false;
			Print(failHeader + " the " + name + ".\n");
		    fi;
		fi;
	    elif st = continue then
		Print(Capitalize(name));
		Print(ambig);
		ok := false;
		done := true;
	    else
		object := FindResult();
		name := FormatName(object@p_oName);
	    fi;
	elif st = continue then
	    Print(Capitalize(name));
	    Print(ambig);
	    ok := false;
	    done := true;
	else
	    object := FindResult();
	    name := FormatName(object@p_oName);
	fi;
	if object = nil and not done then
	    Print(IsAre("There", "no", name, "here.\n"));
	    done := true;
	    ok := false;
	fi;
    fi;
    if done then
	ok
    else
	commonVerbTail(direct, indirect, actorCheck,
		       object, failHeader, verbName, name)
    fi
corp;


/* set up 'SysAdmin' */

CharacterThing(SysAdmin)@p_pDesc :=
    "SysAdmin is the mighty creator of the entire known universe. "
    "His least whim is law. "
    "Nothing is beyond his power. "
    "Beware lest you antagonize him!"$
CharacterThing(SysAdmin)@p_pMoney := 10000$
CharacterThing(SysAdmin)@p_pPrivileged := true$
/* do this right away so we can add checkers as they are defined */
CharacterThing(SysAdmin)@p_pEnterActions := CreateActionList()$
CharacterThing(SysAdmin)@p_pExitActions := CreateActionList()$
CharacterThing(SysAdmin)@p_Image := "Characters/SysAdmin"$


unuse tp_baseUtil
